/***************************************************************************
                           route.cpp
                           version 0.5.5 $Revision: 1.80 $
                           -------------------------------
    copyright            : (C) 2004-2009 by Guido Scholz
    email                : guido.scholz@bayernline.de
    last modified        : $Date: 2009-03-12 16:48:56 $
****************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
   This code implements the route object.
 ***************************************************************************/

#include "element.h"
#include "preferences.h"
#include "route.h"
#include "routedialog.h"

/*some magic strings for reading and writing route data to files*/
static const char RF_ID[]           = "id";
static const char RF_NAME[]         = "name";
static const char RF_TRAIN[]        = "train";
static const char RF_LENGTH[]       = "length";
static const char RF_SPEEDLIMIT[]   = "speedlimit";
static const char RF_ENTRYSIGNAL[]  = "from signal";
static const char RF_EXITSIGNAL[]   = "to signal";
static const char RF_ENTRYTRACK[]   = "entrytrack";
static const char RF_EXITTRACK[]    = "exittrack";
static const char RF_TNDISPLAY[]    = "trainnumberdisplay";
static const char RF_SWITCHXTOY[]   = "switch x to y";
static const char RF_ACTIVATEPORT[] = "activate port";
static const char RF_RELEASEPORT[]  = "release port";
static const char RF_TYPE[]         = "type";


enum trainDirection {tdirNone = 0, tdirBackward, tdirForward,
        tdirEmergency, tdirToggleDir};


Route::Route(unsigned int anid,
        RouteType arouteType,
        const QString& aName,
        const stateElement& aexitSignal,
        const stateElement& aentrySignal,
        const stateElement& tnDisplay,
        const PortState& arePort,
        const PortState& aacPort,
        const QList<stateElement*>& swis,
        QObject* parent, const char* name):
    Section(aName, anid, 0, parent, name),
    routeType(arouteType),
    routestate(rsUnlocked),
    turnouts(0),
    tocounter(0),
    triggerto(NULL),
    replyid(0),
    entrytrack(0),
    exittrack(0),
    exitSignal(aexitSignal),
    entrySignal(aentrySignal),
    rePort(arePort),
    acPort(aacPort)
{
    trainNumberDisplay = tnDisplay;

    /*copy switchitem list*/

    QList<stateElement*>::const_iterator it;
    for (it = swis.constBegin(); it != swis.constEnd(); ++it) {

        stateElement* el = new stateElement;
        if (el != NULL) {
            el->name = (*it)->name;
            el->bus = (*it)->bus;
            el->address = (*it)->address;
            el->state = (*it)->state;
            el->elemPtr = (*it)->elemPtr;

            switchItems.append(el);
        }
    }
}


Route::Route(element* startEl, QObject* parent, const char* name):
    Section(tr("New route"), 0, 0, parent, name),
    routeType(rtRZS),
    routestate(rsUnlocked),
    turnouts(0),
    tocounter(0),
    triggerto(NULL),
    replyid(0),
    entrytrack(0),
    exittrack(0)
{
    exitSignal.state = 0;
    exitSignal.address = 0;
    exitSignal.bus = 1;
    exitSignal.name = "";
    exitSignal.elemPtr = NULL;

    trainNumberDisplay.state = 0;
    trainNumberDisplay.address = 0;
    trainNumberDisplay.bus = 1;
    trainNumberDisplay.name = "";
    trainNumberDisplay.elemPtr = NULL;

    rePort.used = false;
    rePort.switchtooff = false;
    rePort.bus = 1;
    rePort.address = 0;

    acPort.used = false;
    acPort.switchtooff = false;
    acPort.bus = 1;
    acPort.address = 0;

    if (startEl != NULL) {
        sectionName.append(startEl->getLabelText());
        startEl->getStateData(entrySignal);
        //select route type element name dependent
        if (startEl->hasShuntingRouteButtonOnly())
            routeType = rtRRS;
    }
    else {
        entrySignal.state = 0;
        entrySignal.address = 0;
        entrySignal.bus = 1;
        entrySignal.name = "";
        entrySignal.elemPtr = NULL;
    }
}


Route::Route(QTextStream& ts, QObject* parent, const char* name):
    Section(tr("noname"), 0, 0, parent, name),
    routeType(rtRZS),
    routestate(rsUnlocked),
    turnouts(0),
    tocounter(0),
    triggerto(NULL),
    replyid(0),
    entrytrack(0),
    exittrack(0)
{
    /*exit signals are red by default*/
    exitSignal.state = 0;
    exitSignal.name = "";
    exitSignal.elemPtr = NULL;

    entrySignal.name = "";
    entrySignal.elemPtr = NULL;

    trainNumberDisplay.state = 0;
    trainNumberDisplay.address = 0;
    trainNumberDisplay.bus = 1;
    trainNumberDisplay.name = "";
    trainNumberDisplay.elemPtr = NULL;

    readFileTextFromStream(ts);
    updateTrainNumberDisplay();
}


Route::Route(const QString& aName, QObject* parent, const char* name):
    Section(aName, 0, 0, parent, name),
    routeType(rtRZS),
    routestate(rsUnlocked),
    turnouts(0),
    tocounter(0),
    triggerto(NULL),
    replyid(0),
    entrytrack(0),
    exittrack(0)
{
    exitSignal.name = "";
    exitSignal.bus = 1;
    exitSignal.address = 0;
    exitSignal.state = 0;
    exitSignal.elemPtr = NULL;

    entrySignal.name = "";
    entrySignal.bus = 1;
    entrySignal.address = 0;
    entrySignal.state = 0;
    entrySignal.elemPtr = NULL;

    trainNumberDisplay.name = "";
    trainNumberDisplay.bus = 1;
    trainNumberDisplay.address = 0;
    trainNumberDisplay.state = 0;
    trainNumberDisplay.elemPtr = NULL;

    rePort.used = false;
    rePort.switchtooff = false;
    rePort.bus = 1;
    rePort.address = 0;

    acPort.used = false;
    acPort.switchtooff = false;
    acPort.bus = 1;
    acPort.address = 0;
}


Route::~Route()
{
    /* delete each list item*/
    while (!switchItems.isEmpty())
        delete switchItems.takeFirst();
}

/*
 * Update route element names and pointers when layout was edited
 */
void Route::updateElementList(QVector<SpdrPanel*>* elements)
{
    if (elements == NULL)
        return;

    // first clear all element names and pointers
    entrySignal.name = tr("Error");
    entrySignal.elemPtr = NULL;

    exitSignal.name = tr("Error");
    exitSignal.elemPtr = NULL;

    trainNumberDisplay.name = tr("Error");
    trainNumberDisplay.elemPtr = NULL;

    QList<stateElement*>::iterator it;
    for (it = switchItems.begin(); it != switchItems.end(); ++it) {
        (*it)->name = tr("Error");
        (*it)->elemPtr = NULL;
    }

    // now assign the right values
    setupElementList(elements);
}


void Route::setupElementList(QVector<SpdrPanel*>* elements)
{
    if (elements == NULL)
        return;

    for (int i = 0; i < elements->size(); i++) {
        element* el = dynamic_cast<element*>(elements->at(i));

        if (el == NULL)
            continue;

        /*first add switchable elements "between" entry and exit signals*/
        QList<stateElement*>::iterator it;
        for (it = switchItems.begin(); it != switchItems.end(); ++it) {
            if (el->hasSameAddress((*it)->bus, (*it)->address)) {
                (*it)->name = el->getLabelText();
                (*it)->elemPtr = el;
                break;
            }
        }

        /*add exit signal*/
        // FIXME: this finds also Vr-types with same address, check also
        // route start/ end point
        if (el->isRouteMark() && el->hasSameAddress(exitSignal.bus,
                    exitSignal.address)) {
            exitSignal.name = el->getLabelText();
            exitSignal.elemPtr = el;
        }

        /*add entry signal*/
        // FIXME: this finds also Vr-types with same address
        if (el->isRouteMark() && el->hasSameAddress(entrySignal.bus,
                    entrySignal.address)) {
            entrySignal.name = el->getLabelText();
            entrySignal.elemPtr = el;
        }

        /*add train number dislay*/
        if (el->hasSameAddress(trainNumberDisplay.bus,
                    trainNumberDisplay.address)) {
            /*
            fprintf(stderr, "Route: %s, Train: %d, Display: %d %d\n",
                    sectionName.data(), trainid, trainNumberDisplay.bus,
                    trainNumberDisplay.address);
                    */
            trainNumberDisplay.elemPtr = el;
        }
    }

    // update only displays of utilized routes
    if (trainid != 0)
        updateTrainNumberDisplay();
}


void Route::readFileTextFromStream(QTextStream& ts)
{
    QString s, key;

    while (!ts.atEnd()) {
        s = ts.readLine();
        if (!s.startsWith("#")) {
            key = s.section(DS, 0, 0);
            /* key/value pairs are read sequence independent */
            if (key.compare(RF_ID) == 0){
                sectionid = s.section(DS, 1, 1).toUInt();
            }
            else if (key.compare(RF_NAME) == 0){
                sectionName = s.section(DS, 1, 1);
            }
            else if (key.compare(RF_TRAIN) == 0){
                trainid = s.section(DS, 1, 1).toUInt();
            }
            else if (key.compare(RF_LENGTH) == 0){
                length = s.section(DS, 1, 1).toUInt();
            }
            else if (key.compare(RF_SPEEDLIMIT) == 0){
                speedlimit = s.section(DS, 1, 1).toUInt();
            }
            else if (key.compare(RF_EXITSIGNAL) == 0){
                exitSignal.bus = s.section(DS, 1, 1).toUInt();
                exitSignal.address = s.section(DS, 2, 2).toUInt();
            }
            else if (key.compare(RF_EXITTRACK) == 0){
                exittrack = s.section(DS, 1, 1).toUInt();
            }
            else if (key.compare(RF_SWITCHXTOY) == 0){
                stateElement *switchElement = new stateElement;

                if (switchElement != NULL) {
                    switchElement->bus = s.section(DS, 1, 1).toUInt();
                    switchElement->address = s.section(DS, 2, 2).toUInt();
                    switchElement->state = s.section(DS, 3, 3).toUInt();
                    switchElement->elemPtr = NULL;
                    switchElement->name = "";

                    switchItems.append(switchElement);
                }
                // else no memory available
            }
            else if (key.compare(RF_ENTRYSIGNAL) == 0){
                entrySignal.bus = s.section(DS, 1, 1).toUInt();
                entrySignal.address = s.section(DS, 2, 2).toUInt();
                entrySignal.state = s.section(DS, 3, 3).toUInt();
            }
            else if (key.compare(RF_ENTRYTRACK) == 0){
                entrytrack = s.section(DS, 1, 1).toUInt();
            }
            else if (key.compare(RF_TNDISPLAY) == 0){
                trainNumberDisplay.bus = s.section(DS, 1, 1).toUInt();
                trainNumberDisplay.address = s.section(DS, 2, 2).toUInt();
                //trainNumberDisplay.state = s.section(DS, 3, 3).toUInt();
            }
            else if (key.compare(RF_RELEASEPORT) == 0){
                rePort.bus = s.section(DS, 1, 1).toUInt();
                rePort.address = s.section(DS, 2, 2).toUInt();
                rePort.used = (s.section(DS, 3, 3).toInt() == 1);
                rePort.switchtooff = (s.section(DS, 4, 4).toInt() == 1);
            }
            else if (key.compare(RF_ACTIVATEPORT) == 0){
                acPort.bus = s.section(DS, 1, 1).toUInt();
                acPort.address = s.section(DS, 2, 2).toUInt();
                acPort.used = (s.section(DS, 3, 3).toInt() == 1);
                acPort.switchtooff = (s.section(DS, 4, 4).toInt() == 1);
            }
            else if (key.compare(RF_TYPE) == 0){
                routeType = (RouteType)s.section(DS, 1, 1).toUInt();
            }
            /*end of route data*/
            else if (s.startsWith("%%"))
                break;
        }
    }
}


void Route::writeFileTextToStream(QTextStream& ts)
{
    ts  << RF_ID << DS << sectionid << endl
        << RF_NAME << DS << sectionName << endl
        << RF_TRAIN << DS << trainid << endl
        << RF_LENGTH << DS << length << endl
        << RF_SPEEDLIMIT << DS << speedlimit << endl
        << RF_EXITSIGNAL << DS << exitSignal.bus << DS
        << exitSignal.address << endl
        << RF_EXITTRACK << DS << exittrack << endl
        << RF_ENTRYSIGNAL << DS << entrySignal.bus << DS
        << entrySignal.address << DS << entrySignal.state << endl
        << RF_ENTRYTRACK << DS << entrytrack << endl;
    if (trainNumberDisplay.address > 0)
        ts  << RF_TNDISPLAY << DS << trainNumberDisplay.bus << DS
            << trainNumberDisplay.address/* << DS << trainNumberDisplay.state*/
            << endl;
    ts  << RF_RELEASEPORT << DS << rePort.bus << DS << rePort.address
        << DS << rePort.used << DS << rePort.switchtooff << endl
        << RF_ACTIVATEPORT << DS << acPort.bus << DS << acPort.address
        << DS << acPort.used << DS << acPort.switchtooff << endl
        << RF_TYPE << DS << routeType << endl;

    QList<stateElement*>::const_iterator it;
    for (it = switchItems.constBegin(); it != switchItems.constEnd(); ++it) {
        ts
            << RF_SWITCHXTOY << DS
            << (*it)->bus << DS
            << (*it)->address << DS
            << (*it)->state << endl;
    }
}


Route* Route::getClone()
{
    return new Route(sectionid, routeType, sectionName, exitSignal,
            entrySignal, trainNumberDisplay, rePort, acPort,
            switchItems, parent(), "clonedRoute");
}


/*function to update route state visualisation*/
void Route::setState(Route::RouteState state)
{
    if (routestate != state) {
        RouteSetAction rsa = rsaReset;
        routestate = state;

        if (routestate != rsUnlocked) {
            if (routeType == rtRRS || routeType == rtURS)
                rsa = rsaRfs;
            else
                rsa = rsaZfs;
        }

        emit updateRoutePathLEDs(entrySignal, exitSignal, rsa);
        emit stateChanged(this, routestate);
    }
}


int Route::getState()
{
    return routestate;
}


unsigned int Route::getEntryTrack()
{
    return entrytrack;
}


unsigned int Route::getExitTrack()
{
    return exittrack;
}


QString Route::getFromSignalName() const
{
    if (entrySignal.name.isEmpty())
        return QString::number(entrySignal.address);
    else
        return entrySignal.name;
}


QString Route::getToSignalName() const
{
    if (exitSignal.name.isEmpty())
        return QString::number(exitSignal.address);
    else
        return exitSignal.name;
}


void Route::setType(RouteType t)
{
    if (routeType != t)
        routeType = t;
}


int Route::getType() const
{
    return routeType;
}


QString Route::getTypeStr() const
{
    QString typeStr;

    /*returns decoded route type */
    switch (routeType){
        case rtRZS:
            typeStr = QString(QObject::tr("NR")); // normal route
            break;
        case rtUZS:                               // detour route
            typeStr = QString(QObject::tr("DR"));
            break;
        case rtZHS:
            typeStr = QString(QObject::tr("HR")); // help route
            break;
        case rtRRS:
            typeStr = QString(QObject::tr("NS")); // normal shunting
            break;
        case rtURS:                               // detour shunting
            typeStr = QString(QObject::tr("DS"));
            break;
    }
    return typeStr;
}

/**
 * Start an unlocked route. This function returns tree differend values,
 * depending on processing results.
 *   -2: routing is interrupted because a turnout to switch is occupied
 *   -1: routing is interrupted because nonshunting route leads over
         occupied elements
 *    0: routing is interrupted due to a locked turnout/signal
 *    1: routing is OK
 */
int Route::startRouting()
{
    /*No valid entry signal*/
    if (entrySignal.elemPtr == NULL)
        return -3;

    /*No valid exit signal*/
    if (exitSignal.elemPtr == NULL)
        return -3;

    /*
     * 1) check for locked elements
     */
    if (entrySignal.elemPtr->isLocked())
        return 0;

    // this case may lead to a route reservation
    if (exitSignal.elemPtr->isLocked() &&
            exitSignal.elemPtr->hasDifferentState(exitSignal.state)) {
        routestate = rsWfLock;
        emit stateChanged(this, routestate);
        return 0;
    }

    turnouts = 0;
    tocounter = 0;

    QList<stateElement*>::const_iterator it;
    for (it = switchItems.constBegin(); it != switchItems.constEnd(); ++it) {
        element* el = (*it)->elemPtr;
        if (el != NULL && el->isLocked() &&
                (el->hasDifferentState((*it)->state) ||
                 el->is2StateDKW() || !el->isLockable())) {
            routestate = rsWfLock;
            emit stateChanged(this, routestate);
            return 0;
        }

        // count turnouts for timer activation
        if (el != NULL && el->isTurnout() &&
                el->hasDifferentState((*it)->state)) {
            ++turnouts;
            el->setSwitched(false);
            if (el->isOccupied())
                return -2;
        }
    }

    //fprintf(stderr, "counter: %d  turnouts: %d\n", tocounter, turnouts);

    /*
     * 2) switch route elements but without locking
     */
    if (exitSignal.elemPtr != NULL) {
           exitSignal.elemPtr->switchToDir(exitSignal.state);
           exitSignal.elemPtr->repaint();
    }

    // start timer controlled turnout switching
    switchTurnouts();
    return 1;
}

/*
 * Timer controlled loop over all turnouts.
 * The original SpDr waits 250 ms until next turnout is
 * switched to avoid high power consumption. Signals on
 * route path are switched after "Fahrstrassenfestlegemelder".
 * Option to force turnout switching: prf.sendstate
 */
void Route::switchTurnouts()
{
    if (turnouts > 0) {
        QList<stateElement*>::const_iterator it;
        for (it = switchItems.constBegin(); it != switchItems.constEnd(); ++it) {
            element* el = (*it)->elemPtr;

            if (el != NULL && el->isTurnout() && !el->isSwitched()) {
                if (el->hasDifferentState((*it)->state)) {

                    ++tocounter;
                    // last turnout will trigger route path highlighting
                    if (tocounter == turnouts) {
                        triggerto = el;
                        connect(el, SIGNAL(turnoutIsSwitched()),
                                this, SLOT(showRoutePath()));
                    }

                    el->switchToDir((*it)->state);
                    el->repaint();
                    el->setSwitched(true);

                    //fprintf(stderr, "counter: %d  turnouts: %d\n",
                    //        tocounter, turnouts);
                    if (tocounter < turnouts) {
                        QTimer::singleShot(250, this, SLOT(switchTurnouts()));
                    }
                }
                // no blink animation but command sending necessary
                else if (pref.sendstate)
                    el->sendSrcpState();
            }
        }
    }
    // if there is no single turnout go ahead anyway
    else
        showRoutePath();
}


void Route::showRoutePath()
{
    // clean up
    if (triggerto != NULL) {
        disconnect(triggerto, SIGNAL(turnoutIsSwitched()),
                this, SLOT(showRoutePath()));
        triggerto = NULL;
    }

    /*
     * 3) switch route path element LEDs to yellow,
     *    check for occupied elements if not shunting route
     */
    RouteSetAction rsa = rsaZfs;
    if (routeType == rtRRS || routeType == rtURS) {
        rsa = rsaRfs;
    }
    // send signal to gbs to change route path LEDs
    emit updateRoutePathLEDs(entrySignal, exitSignal, rsa);

    // Interrupt routing if "Zugfahrstrasse" meets occupied element.
    // The real lock state should be reached when all
    // occupied segments changed to be unoccupied.
    // route states: rsUnlocked, rsLocked, rsWfLock, rsWfUnlock
    // occupied tracks go this ways:
    //   rsUnlocked -> rsWfLock -> rsLocked
    //   rsLocked -> rsWfUnlock -> rsUnlocked
    if (rsaReset == rsa) {
        routestate = rsWfLock;
        emit stateChanged(this, routestate);
        return;
    }

    /*
     * FIXME:
     * 4) lock all switchable elements; in original SpDr this is done
     * with step 3) but is too complicated to implement respecting
     * interruption by occupied elements for "Zugfahrstrassen" and the
     * necessary unlocking of already locked elements
     */
    if (entrySignal.elemPtr != NULL)
        entrySignal.elemPtr->setLocked(true);

    /*exit signal does not need locking*/

    QList<stateElement*>::const_iterator it;
    for (it = switchItems.constBegin(); it != switchItems.constEnd(); ++it) {
        element* el = (*it)->elemPtr;
        if (el != NULL)
            el->setLocked(true);
    }

    /*
     * 5) activate "Fahrstrassenfestlegemelder" (FfM) at entry signal,
     *    shunting routes do not have an active FfM
     */
    if (entrySignal.elemPtr != NULL)
           entrySignal.elemPtr->activateFfM((routeType != rtRRS &&
                       routeType != rtURS));

    /*
     * 6) switch signals on route path to Sh1 (Siemens Type)
     * without changing lock state, also simple GA elements are switched
     * now
     */
    for (it = switchItems.constBegin(); it != switchItems.constEnd(); ++it) {
        element* el = (*it)->elemPtr;
        if (el != NULL)
            if (el->isSignal() || el->isSimpleGA()) {
                el->switchToDir((*it)->state);
                el->repaint();
            }
    }

    /* 7) at last switch entry signal to Hp1/Sh1 etc.*/
    if (entrySignal.elemPtr != NULL)
        entrySignal.elemPtr->switchToDir(entrySignal.state);

    routestate = rsLocked;
    emit stateChanged(this, routestate);
}


/**
 * Stop an active route. All signals are switched to red light,
 * turnouts keep current direction, a FfM is deactivated, locked
 * elements are unlocked.
 */
void Route::stopRouting()
{
    if (entrySignal.elemPtr != NULL) {
        entrySignal.elemPtr->activateFfM(false);
        entrySignal.elemPtr->switchToDir(0);
        //FIXME: temporary solution
        if (routestate == rsLocked)
            entrySignal.elemPtr->setLocked(false);
    }

    QList<stateElement*>::const_iterator it;
    for (it = switchItems.constBegin(); it != switchItems.constEnd(); ++it) {
        element* el = (*it)->elemPtr;
        if (el != NULL) {
            if (el->isSignal())
                el->switchToDir(0);
            //FIXME: temporary solution
            if (routestate == rsLocked)
                el->setLocked(false);
        }
    }

    /*
     * update route path element LEDs and
     * send signal to gbs to change route path LEDs
     */
    RouteSetAction rsa = rsaReset;
    emit updateRoutePathLEDs(entrySignal, exitSignal, rsa);

    routestate = rsUnlocked;
    emit stateChanged(this, routestate);
}


void Route::hideRoute()
{
    if (exitSignal.elemPtr != NULL)
        exitSignal.elemPtr->switchSelectionMode(ksmNormal);

    if (entrySignal.elemPtr != NULL) {
        entrySignal.elemPtr->switchToDir(0);
        entrySignal.elemPtr->switchSelectionMode(ksmNormal);
    }

    if (trainNumberDisplay.elemPtr != NULL) {
        //FIXME: check trainNumberDisplay.elemPtr->updateTrainNumber(0);
        trainNumberDisplay.elemPtr->switchSelectionMode(ksmNormal);
    }

    QList<stateElement*>::const_iterator it;
    for (it = switchItems.constBegin(); it != switchItems.constEnd(); ++it) {
        if ((*it)->elemPtr != NULL) {
            (*it)->elemPtr->switchSelectionMode(ksmNormal);
            if ((*it)->elemPtr->isSignal())
                (*it)->elemPtr->switchToDir(0);
        }
    }
}


void Route::showRoute()
{
    if (exitSignal.elemPtr != NULL)
        exitSignal.elemPtr->showElementState(exitSignal.state,
                ksmStopSig);

    if (entrySignal.elemPtr != NULL)
        entrySignal.elemPtr->showElementState(entrySignal.state,
                ksmStartSig);

    if (trainNumberDisplay.elemPtr != NULL) {
        trainNumberDisplay.elemPtr->updateTrainNumber(trainid);
        trainNumberDisplay.elemPtr->showElementState(-1, ksmDisplay);
    }

    QList<stateElement*>::const_iterator it;
    for (it = switchItems.constBegin(); it != switchItems.constEnd(); ++it) {
        if ((*it)->elemPtr != NULL)
            (*it)->elemPtr->showElementState((*it)->state, ksmSwitchEl);
    }
}


void Route::setEntrySignal(element* el)
{
    el->getStateData(entrySignal);
    el->switchSelectionMode(ksmStartSig);
    updateRouteName();
    updateRouteType();
}


void Route::setExitSignal(element* el)
{
    el->getStateData(exitSignal);
    el->switchSelectionMode(ksmStopSig);
    updateRouteName();
    updateRouteType();
}


void Route::updateRouteName()
{
    if (hasEntrySignal()) {
        if (hasExitSignal())
            sectionName = QString("%1 - %2").arg(entrySignal.name)
                .arg(exitSignal.name);
        else
            sectionName = QString(tr("New route from %1")
                    .arg(entrySignal.name));
    }
    else {
        if (hasExitSignal())
            sectionName = QString(tr("New route to %1").arg(exitSignal.name));
        else
            sectionName = tr("New route");
    }
}

/*
 * In route edit mode try to guess what type of route is recorded,
 * but only respecting changes from default rtRZS to new rtRRS.
 */
void Route::updateRouteType()
{
    if (rtRZS != routeType)
        return;

    if (hasEntrySignal() &&
            entrySignal.elemPtr->hasShuntingRouteButtonOnly()) {
        routeType = rtRRS;
        return;
    }

    if (hasExitSignal() &&
            exitSignal.elemPtr->hasShuntingRouteButtonOnly())
        routeType = rtRRS;
}


void Route::addSwitchElement(element* el)
{
    stateElement* se = new stateElement;
    if (se != NULL) {
        el->getStateData(*se);
        switchItems.append(se);
        el->switchSelectionMode(ksmSwitchEl);
    }
    // else no memory available
}


void Route::removeElement(element* el)
{
    elemSelectionMode sm = el->getSelectionMode();
    switch (sm) {

        case (ksmStartSig):
            entrySignal.state = 0;
            entrySignal.address = 0;
            entrySignal.bus = 0;
            entrySignal.name = "";
            entrySignal.elemPtr = NULL;
            updateRouteName();
            break;

        case (ksmStopSig):
            exitSignal.state = 0;
            exitSignal.address = 0;
            exitSignal.bus = 0;
            exitSignal.name = "";
            exitSignal.elemPtr = NULL;
            updateRouteName();
            break;

        case (ksmDisplay):
            el->updateTrainNumber(0);
            trainNumberDisplay.state = 0;
            trainNumberDisplay.address = 0;
            trainNumberDisplay.bus = 0;
            trainNumberDisplay.name = "";
            trainNumberDisplay.elemPtr = NULL;
            break;

        case (ksmSwitchEl):
            {
                bool found = false;

                QList<stateElement*>::const_iterator it;
                for (it = switchItems.constBegin();
                        it != switchItems.constEnd(); ++it) {
                    if ((*it)->elemPtr == el) {
                        found = true;
                        break;
                    }
                }
                if (found) {
                    /*remove this item from list*/
                    int i = switchItems.indexOf(*it);
                    if (i != -1)
                        delete switchItems.takeAt(i);
                }
                break;
            }
        case (ksmNormal):
            break;

        default:
            break;
    }
    el->switchSelectionMode(ksmNormal);
}


bool Route::hasEntrySignal()
{
    return entrySignal.elemPtr != NULL;
}


bool Route::hasExitSignal()
{
    return exitSignal.elemPtr != NULL;
}


bool Route::hasThisExitSignal(element* el)
{
    return (exitSignal.elemPtr == el);
}


bool Route::isLockedWithEntrySignal(element* el)
{
    return routestate != rsUnlocked &&
        (entrySignal.elemPtr == el);
}


bool Route::isUnlockedWithEntrySignalType(element* el, GbsButtonState cb,
        GbsButtonState sb)
{
    /* Function matrix

       route type    c-button       s-button
       ---------------------------------------
         rtRZS      kZfsClicked   kNoneClicked
         rtUZS      kZfsClicked   kUfgtClicked
         rtZHS      kZhsClicked   kNoneClicked
                    kZfsClicked   kNoneClicked
         rtRRS      kRfsClicked   kNoneClicked
         rtURS      kRfsClicked   kUfgtClicked
       ---------------------------------------
    */
    if (routestate == rsUnlocked && (entrySignal.elemPtr == el)) {

        bool returnvalue = false;
        switch (routeType){
            case rtRZS:
                returnvalue = (kZfsClicked == cb && kNoneClicked == sb);
                break;
            case rtUZS:
                returnvalue = (kZfsClicked == cb && kUfgtClicked == sb);
                break;
            case rtZHS:
                returnvalue = (kZhsClicked == cb || kZfsClicked == cb )
                    && kNoneClicked == sb;
                break;
            case rtRRS:
                returnvalue = (kRfsClicked == cb && kNoneClicked == sb);
                break;
            case rtURS:
                returnvalue = (kRfsClicked == cb && kUfgtClicked == sb);
                break;
            default:
                break;
        }
        return returnvalue;
    }
    else
        return false;
}


bool Route::isUnlockedType(element* fel, element* tel, GbsButtonState cb,
        GbsButtonState sb)
{
    /* Function matrix

       route type    c-button       s-button
       ---------------------------------------
         rtRZS      kZfsClicked   kNoneClicked
         rtUZS      kZfsClicked   kUfgtClicked
         rtZHS      kZhsClicked   kNoneClicked
                    kZfsClicked   kNoneClicked
         rtRRS      kRfsClicked   kNoneClicked
         rtURS      kRfsClicked   kUfgtClicked
       ---------------------------------------
    */
    if (routestate == rsUnlocked && (entrySignal.elemPtr == fel) &&
            (exitSignal.elemPtr == tel)) {

        bool returnvalue = false;
        switch (routeType){
            case rtRZS:
                returnvalue = (kZfsClicked == cb && kNoneClicked == sb);
                break;
            case rtUZS:
                returnvalue = (kZfsClicked == cb && kUfgtClicked == sb);
                break;
            case rtZHS:
                returnvalue = (kZhsClicked == cb || kZfsClicked == cb )
                    && kNoneClicked == sb;
                break;
            case rtRRS:
                returnvalue = (kRfsClicked == cb && kNoneClicked == sb);
                break;
            case rtURS:
                returnvalue = (kRfsClicked == cb && kUfgtClicked == sb);
                break;
            default:
                break;
        }
        return returnvalue;
    }
    else
        return false;
}


/*
  activate route action if conditions are given:

  used  switchtooff  ison  action
  -------------------------------
   1         0        0      0
   1         0        1      1
   1         1        0      1
   1         1        1      0
  -------------------------------
 */
bool Route::canActivateByFeedbackPort(unsigned int bus,
        unsigned int port, bool ison)
{
    return (routestate == rsUnlocked && acPort.used && acPort.bus == bus &&
        acPort.address == port && acPort.switchtooff != ison);
}


bool Route::canReleaseByFeedbackPort(unsigned int bus,
        unsigned int port, bool ison)
{
    return (routestate != rsUnlocked && rePort.used && rePort.bus == bus &&
        rePort.address == port && rePort.switchtooff != ison);
}


bool Route::runEditDialog(QWidget* dlgparent)
{
    bool returnvalue = false;

    RouteDialog* rtDlg = new RouteDialog(dlgparent);

    if (rtDlg == NULL)
        return returnvalue;

    connect(rtDlg, SIGNAL(getElementByAddress(const int, const int,
                    element**)), this,
            SIGNAL(getElementByAddress(const int, const int,
                    element**)));
    rtDlg->setRouteNumber(sectionid);
    rtDlg->setRouteTrain(trainid);
    rtDlg->setRouteLength(length);
    rtDlg->setRouteSpeedLimit(speedlimit);
    rtDlg->setRouteName(sectionName);
    rtDlg->setEntrySignalData(entrySignal);
    rtDlg->setExitSignalData(exitSignal);
    rtDlg->setSignalTracks(entrytrack, exittrack);
    rtDlg->setActivateData(acPort);
    rtDlg->setReleaseData(rePort);
    rtDlg->setRouteType(routeType);
    rtDlg->setRouteElements(switchItems);
    if (rtDlg->exec() == QDialog::Accepted) {
        // update gbs: 1) hide old route 2) show new route
        hideRoute();
        sectionid = rtDlg->getRouteNumber();
        setTrain(rtDlg->getRouteTrain());
        length = rtDlg->getRouteLength();
        speedlimit = rtDlg->getRouteSpeedLimit();
        sectionName = rtDlg->getRouteName();
        rtDlg->getEntrySignalData(entrySignal);
        rtDlg->getExitSignalData(exitSignal);
        rtDlg->getSignalTracks(entrytrack, exittrack);
        rtDlg->getActivateData(acPort);
        rtDlg->getReleaseData(rePort);
        routeType = (RouteType) rtDlg->getRouteType();
        /* delete each list item*/
        while (!switchItems.isEmpty())
            delete switchItems.takeFirst();
        rtDlg->getRouteElements(switchItems);

        showRoute();
        returnvalue = true;
    }
    disconnect(rtDlg, SIGNAL(getElementByAddress(const int,
                    const int, element**)), this,
            SIGNAL(getElementByAddress(const int, const int,
                    element**)));
    delete rtDlg;
    return returnvalue;
}

/*
 * return pointer to entry signal element
 */
element* Route::getEntrySignalElementPtr()
{
    return entrySignal.elemPtr;
}

/* return pointer to exit signal element */
element* Route::getExitSignalElementPtr()
{
    return exitSignal.elemPtr;
}

/*assemble CRCF route info message string*/
QString Route::getCrcfInfoMessage(CrcfMessage::CrcfAttribute at) const
{
    unsigned int result = 0;

    switch (at) {
        case CrcfMessage::atId:
            result = sectionid;
            break;
        case CrcfMessage::atName:
            return CrcfMessage::message(CrcfMessage::acRoute, sectionid,
                    CrcfMessage::meInfo, at, sectionName);
        case CrcfMessage::atState:
            /*TODO: add other states*/
            result = (routestate == rsUnlocked) ? 0 : 1;
            break;
        case CrcfMessage::atType:
            result = routeType;
            break;
        case CrcfMessage::atTrain:
            result = trainid;
            break;
        case CrcfMessage::atDirection:
            // special message about train direction change
            // should be called in: getCrcfSetMessage()
            return CrcfMessage::message(CrcfMessage::acTrain, trainid,
                    CrcfMessage::meSet, at, tdirToggleDir);
        case CrcfMessage::atLength:
            result = length;
            break;
        case CrcfMessage::atSpeedLimit:
            result = speedlimit;
            break;
        case CrcfMessage::atEntrySignalAspect:
            {
            stateElement se;
            entrySignal.elemPtr->getStateData(se);
            result = se.state + 1;
            }
            break;
        case CrcfMessage::atExitSignalAspect:
            {
            stateElement se;
            exitSignal.elemPtr->getStateData(se);
            result = se.state + 1;
            }
            break;
        default:
            return "";
            break;
    }

    return CrcfMessage::message(CrcfMessage::acRoute, sectionid,
            CrcfMessage::meInfo, at, result);
}

void Route::setCrcfMessageReplyId(unsigned int id)
{
    replyid = id;
}

unsigned int Route::getCrcfMessageReplyId()
{
    return replyid;
}

bool Route::isCrcfMessageRequested()
{
    return (replyid != 0);
}

