/***************************************************************************
                           router.cpp
                           version 0.5.3 $Revision: 1.75 $
                           -------------------------------
    Copyright            : (C) 2004-2009 by Guido Scholz
    email                : guido.scholz@bayernline.de
    last modified        : $Date: 2009-03-16 17:38:04 $
****************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
  This code implements the router object. This object cares for the route
  list, loads it from file, save it to file, shows route edit window etc.
 ***************************************************************************/

#include <QApplication>
#include <QFile>
#include <QDateTime>

#include "element.h"
#include "preferences.h"
#include "resources.h"
#include "router.h"


Router::Router(QObject* parent, const char* name):
    QObject(parent), selectedRoute(NULL), resetRt(NULL),
    selectedStartSig(NULL), modified(false), serverhasgm(false),
    visualmode(kvmNormal), lastcb(kNoneClicked), infosessionid(0),
    gbsElements(NULL)
{
    setObjectName(name);
}


Router::Router(QObject* parent, QVector<SpdrPanel*>* elPtr, const char* name):
    QObject(parent), selectedRoute(NULL), resetRt(NULL),
    selectedStartSig(NULL), modified(false), serverhasgm(false),
    visualmode(kvmNormal), lastcb(kNoneClicked), infosessionid(0),
    gbsElements(elPtr)
{
    setObjectName(name);
}


Router::~Router()
{
    /* delete each list item*/
    while (!routeList.isEmpty())
        delete routeList.takeFirst();
}

/*
 * read a new route from file
 * set an id if the route does not have one already
 */
void Router::readFileTextFromStream(QTextStream& ts)
{
    Route* rt = NULL;

    /*clear old list*/
    while (!routeList.isEmpty())
        delete routeList.takeFirst();

    while (!ts.atEnd()) {
        rt = new Route(ts, this, "streamRoute");
        if (rt != NULL) {
            if (rt->getId() == 0)
                rt->setId(getMaximumRouteIdNumber() + 1);
            routeList.append(rt);
        }
    }

    modified = false;
    setupRouteElements();
    /*send update signal to routing viewer*/
    emit routeListChanged();
}

void Router::setupRouteElements()
{
    if (gbsElements == NULL)
        return;

    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        (*it)->setupElementList(gbsElements);
        connect((*it), SIGNAL(stateChanged(Route*, int)),
                this, SLOT(processRouteState(Route*, int)));
        connect((*it), SIGNAL(updateRoutePathLEDs(const stateElement&,
                        const stateElement&, Route::RouteSetAction&)),
                this, SIGNAL(updateRoutePathLEDs(const stateElement&,
                        const stateElement&, Route::RouteSetAction&)));
    }
}


void Router::updateRouteElements()
{
    if (gbsElements == NULL)
        return;

    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        (*it)->updateElementList(gbsElements);
        // tell route list window about changed data
        emit routeDataChanged(*it);
    }
}


unsigned int Router::getRouteCount()
{
    return routeList.count();
}


void Router::copyRouteAt(unsigned int index)
{
    copyRoute(getRouteAt(index));
}


Route* Router::copyRoute(Route* cr)
{
    Route* nr = NULL;

    if (cr != NULL) {
        nr = cr->getClone();
        if (nr != NULL) {
            routeList.insert(routeList.indexOf(cr), nr);
            connect(nr, SIGNAL(stateChanged(Route*, int)),
                    this, SLOT(processRouteState(Route*, int)));
            connect(nr, SIGNAL(updateRoutePathLEDs(const stateElement&,
                            const stateElement&, Route::RouteSetAction&)),
                    this, SIGNAL(updateRoutePathLEDs(const stateElement&,
                            const stateElement&, Route::RouteSetAction&)));
            modified = true;
        }
    }
    return nr;
}


void Router::deleteRouteAt(unsigned int index)
{
    deleteRoute(routeList.at(index));
}


void Router::deleteRoute(Route* dr)
{
    if (selectedRoute == dr)
        selectedRoute = NULL;

    if (dr != NULL) {
        if (dr->getState() != Route::rsUnlocked) {
            // this is a withdrawal, no release
            withdrawRoute(dr);
        }
        dr->hideRoute();
        disconnect(dr, SIGNAL(updateRoutePathLEDs(const stateElement&,
                        const stateElement&, Route::RouteSetAction&)),
                this, SIGNAL(updateRoutePathLEDs(const stateElement&,
                        const stateElement&, Route::RouteSetAction&)));
        disconnect(dr, SIGNAL(stateChanged(Route*, int)),
                this, SLOT(processRouteState(Route*, int)));
        int i = routeList.indexOf(dr);
        if (i != -1)
            delete routeList.takeAt(i);
        modified = true;
    }
}


void Router::writeFileTextToStream(QTextStream& ts)
{
    ts
        << "# start of route section" << endl
        << "# routes=" << routeList.count() << endl;

    unsigned int i = 0;

    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        i++;
        ts << "%% route " << i << endl;
        (*it)->writeFileTextToStream(ts);
    }
    modified = false;
}


Route* Router::getRouteAt(unsigned int aRouteNo)
{
    return routeList.at(aRouteNo);
}


bool Router::isModified()
{
    return modified;
}


void Router::clear()
{
    while (!routeList.isEmpty())
        delete routeList.takeFirst();
    modified = false;
}


void Router::clearRoutes()
{
    clear();
    /*send update signal to routingviewer*/
    emit routeListChanged();
}


/**
 * update route element highlighting in route edit mode
 * first hide last selected route then show current selected one
 */
void Router::selectedRouteChanged(Route* currentRoute)
{
    if (visualmode != kvmEditRoute)
        return;

    if (selectedRoute != NULL)
        selectedRoute->hideRoute();

    if (currentRoute != NULL) {
        currentRoute->showRoute();
        selectedRoute = currentRoute;
    }
}


void Router::showRouteAt(int idx)
{
    if (visualmode == kvmEditRoute) {
        Route* cr = getRouteAt(idx);
        if (cr != NULL)
            cr->showRoute();
    }
}


void Router::switchVisualMode(elemVisualMode vm)
{
    // if layout was in edit mode update route data to make sure
    // that changed elements are found at the right place
    if (visualmode == kvmEditLayout)
        updateRouteElements();

    if (visualmode == kvmEditRoute && vm != kvmEditRoute) {
        if (selectedRoute != NULL) {
            selectedRoute->hideRoute();
            selectedRoute = NULL;
        }
    }

    visualmode = vm;
}


void Router::recordElement(element* el, elemRecordType rtype)
{
    if (visualmode == kvmEditRoute && selectedRoute != NULL) {
        switch (rtype) {

            case krecStartStop:
                if (!selectedRoute->hasEntrySignal()) {
                    if (el->showsStop()) {
                        QApplication::beep();
                        emit statusMessage(
                                tr("Error: Start signal '%1' in stop position")
                                .arg(el->getLabelText()));
                    }
                    else {
                        selectedRoute->setEntrySignal(el);

                        /*send update signal to routingviewer to show
                          changed route name*/
                        emit routeDataChanged(selectedRoute);
                        modified = true;
                    }
                }
                else if (!selectedRoute->hasExitSignal()) {
                    selectedRoute->setExitSignal(el);

                    /*send update signal to routingviewer to show changed
                      route name*/
                    emit routeDataChanged(selectedRoute);
                    modified = true;
                }
                else {
                    QApplication::beep();
                    emit statusMessage(
                            tr("Recording of signal '%1' failed")
                            .arg(el->getLabelText()));
                }
                break;

            case krecDisplay:
                if (!selectedRoute->hasTrainNumberDisplay())
                    selectedRoute->setTrainNumberDisplay(el);
                modified = true;
                break;

            case krecNormal:
                selectedRoute->addSwitchElement(el);
                modified = true;
                break;

            case krecClear:
                selectedRoute->removeElement(el);
                /*send update signal to routingviewer to show changed
                  route name if changed element was entry or exit signal*/
                emit routeDataChanged(selectedRoute);
                modified = true;
                break;

            default:
                break;
        }
    }
}


Route* Router::addNewRoute()
{
    Route* nr = new Route(tr("New route"), this, "newRoute");
    if (nr != NULL) {
        nr->setId(getMaximumRouteIdNumber() + 1);
        selectedRouteChanged(nr);
        routeList.append(nr);

        connect(nr, SIGNAL(stateChanged(Route*, int)),
                this, SLOT(processRouteState(Route*, int)));
        connect(nr, SIGNAL(updateRoutePathLEDs(const stateElement&,
                        const stateElement&, Route::RouteSetAction&)),
                this, SIGNAL(updateRoutePathLEDs(const stateElement&,
                        const stateElement&, Route::RouteSetAction&)));
    }
    return nr;
}


void Router::startRecordModeAt(unsigned int index)
{
    selectedRoute = getRouteAt(index);
    showRouteAt(index);
}

/*search for route linked to this one*/
Route* Router::getUtilizedRouteToRoute(Route* tr)
{
    if (tr == NULL)
        return NULL;

    element* signal = tr->getEntrySignalElementPtr();
    if (signal == NULL)
        return NULL;

    Route* sr = getUtilizedRouteToSignal(signal);
    if (sr == NULL) {
        unsigned int track = tr->getEntryTrack();
        if (0 == track)
            return NULL;

        sr = getUtilizedRouteToTrack(track);
        if (sr == NULL)
            return NULL;
    }
    return sr;
}

/*
 * transfer train number from source route to target route and update
 * route data visualization
 */
void Router::transferTrainNumber(Route* tr)
{
    Route* sr = getUtilizedRouteToRoute(tr);
    if (sr == NULL)
        return;

    /* 4: broadcast new exit signal (Hp0) aspect of last route*/
    if (pref.gmbroadcastroutestate) {
        QString cms = sr->getCrcfInfoMessage(CrcfMessage::atExitSignalAspect);
        sendGmCrcfMessage(0, infosessionid, cms);
    }

    tr->setTrain(sr->getTrain());
    emit routeDataChanged(tr);

    sr->clearTrain();
    emit routeDataChanged(sr);
    modified = true;

    /* 5: broadcast new train assignments*/
    if (pref.gmbroadcasttrainnumber) {
        QString cms = sr->getCrcfInfoMessage(CrcfMessage::atTrain);
        sendGmCrcfMessage(0, infosessionid, cms);
        cms = tr->getCrcfInfoMessage(CrcfMessage::atTrain);
        sendGmCrcfMessage(0, infosessionid, cms);
    }
}

/*
 * Routes trigger this method to report state changes.
 * If the route was unlocked no status message is send. This must be
 * done by the caller to differentiate withdrawal and release.
 */
void Router::processRouteState(Route* rt, int rs)
{
    // send signal to route list viewer to update state visualisation
    emit routeStateChanged(rt);

    switch ((Route::RouteState)rs) {
        case Route::rsUnlocked:
            {
            QString cms = rt->getCrcfInfoMessage(CrcfMessage::atState);
            if (pref.gmbroadcastroutestate)
                sendGmCrcfMessage(0, infosessionid, cms);
            else
                if (rt->isCrcfMessageRequested()) {
                    sendGmCrcfMessage(rt->getCrcfMessageReplyId(),
                            infosessionid, cms);
                    rt->setCrcfMessageReplyId(0);
                }

            cms = rt->getCrcfInfoMessage(CrcfMessage::atEntrySignalAspect);
            if (pref.gmbroadcastroutestate)
                sendGmCrcfMessage(0, infosessionid, cms);
            else
                if (rt->isCrcfMessageRequested()) {
                    sendGmCrcfMessage(rt->getCrcfMessageReplyId(),
                            infosessionid, cms);
                    rt->setCrcfMessageReplyId(0);
                }
            }
            break;

        case Route::rsLocked:
            {
            emit statusMessage(tr("Route '%1' activated")
                    .arg(rt->getSectionName()));
            /* 1: send route state message*/
            QString cms = rt->getCrcfInfoMessage(CrcfMessage::atState);
            if (pref.gmbroadcastroutestate)
                sendGmCrcfMessage(0, infosessionid, cms);
            else
                if (rt->isCrcfMessageRequested()) {
                    sendGmCrcfMessage(rt->getCrcfMessageReplyId(),
                            infosessionid, cms);
                    rt->setCrcfMessageReplyId(0);
                }

            if (!pref.gmbroadcastroutestate)
                break;

            //if there is a route which exit signal is the
            //same as this routes entry signal then create CRCF message
            //about exit signal aspect change of respective route
            element* signal = rt->getEntrySignalElementPtr();
            if (signal == NULL)
                break;

            bool sharedsignal = false;
            bool sharedtrack = false;
            // first search route sharing same signal
            Route* sr = getUtilizedRouteToSignal(signal);
            if (sr == NULL) {
                //second try: find route to same track
                unsigned int track = rt->getEntryTrack();
                if (0 == track)
                    break;

                sr = getUtilizedRouteToTrack(track);
                if (sr == NULL)
                    break;
                else
                    sharedtrack = true;
            }
            else
                sharedsignal = true;

            /* 2: send train direction change message*/
            // If entry signal of this route has different direction than
            // exit signal of last route then send train direction
            // change message
            if (sharedsignal) {
                if (signal->routeDirectionsDiffer()) {
                    cms = sr->getCrcfInfoMessage(CrcfMessage::atDirection);
                    sendGmCrcfMessage(0, infosessionid, cms);
                }
                /* 3: send signal aspect (Hp1 etc.) change notification*/
                cms = sr->getCrcfInfoMessage(CrcfMessage::atExitSignalAspect);
                sendGmCrcfMessage(0, infosessionid, cms);
            }
            else if (sharedtrack) {
                element* exitsignal = sr->getExitSignalElementPtr();
                if (NULL == exitsignal)
                    break;

                if (signal->entryDir() != exitsignal->entryDir()) {
                    cms = sr->getCrcfInfoMessage(CrcfMessage::atDirection);
                    sendGmCrcfMessage(0, infosessionid, cms);
                }
            }

            /* 4: send next route id to train*/
            cms = CrcfMessage::message(CrcfMessage::acTrain,
                    sr->getTrain(), CrcfMessage::meSet,
                    CrcfMessage::atNextRoute, rt->getId());
            sendGmCrcfMessage(0, infosessionid, cms);
            }
            break;
        case Route::rsWfLock:
            emit statusMessage(tr("Route '%1' waiting for activation")
                    .arg(rt->getSectionName()));
            break;
        case Route::rsWfUnlock:
            emit statusMessage(tr("Route '%1' waiting for release")
                    .arg(rt->getSectionName()));
            break;
        case Route::rsLocking:
            emit statusMessage(tr("Route '%1' is in activating state")
                    .arg(rt->getSectionName()));
            break;
    }
}


void Router::activateRouteAt(unsigned int index)
{
    Route* rt = getRouteAt(index);
    activateRoute(rt);
}


void Router::activateRoute(Route* rt)
{
    if (rt == NULL || visualmode == kvmEditLayout)
        return;

    int result = rt->startRouting();
    switch (result) {
        case 1:
            //processRouteState(rt, Route::rsLocked);
            break;
        case 0:
            QApplication::beep();
            emit statusMessage(tr("Route '%1' is locked by an other "
                        "route, reservation done.")
                    .arg(rt->getSectionName()));
            break;
        case -1:
            QApplication::beep();
            emit statusMessage(tr("No routing possible; "
                        "route '%1' is blocked by occupied element.")
                    .arg(rt->getSectionName()));
            break;
        case -2:
            QApplication::beep();
            emit statusMessage(tr("No routing possible; "
                        "route '%1' is blocked by occupied turnout.")
                    .arg(rt->getSectionName()));
            break;
        case -3:
            emit statusMessage(tr("Missing signal item pointer."));
            break;
        default:
            emit statusMessage(tr("Unknown routing error."));
            break;
    }
}


void Router::releaseRouteAt(unsigned int index)
{
    Route* rt = getRouteAt(index);
    releaseRoute(rt);
}

/*
 * release the selected route, send a status message and do a train
 * number transfer
 */
void Router::releaseRoute(Route* rt)
{
    if (rt == NULL || visualmode == kvmEditLayout)
        return;

    rt->stopRouting();
    emit statusMessage(tr("Route '%1' released").arg(rt->getSectionName()));
    transferTrainNumber(rt);

    /* Check for reserved route to entry signal of this route and
     * activate if available*/
    Route* nextrt = getReservedRouteWithExitSignal(
            rt->getEntrySignalElementPtr());
    if (NULL == nextrt) {

        /* If not found check for reserved route to exit signal of this
         * route and activate if available*/
        nextrt = getReservedRouteWithExitSignal(
                rt->getExitSignalElementPtr());
        if (NULL == nextrt)
            return;
    }

    emit statusMessage(tr("Reserved route '%1' found to activate")
            .arg(nextrt->getSectionName()));
    activateRoute(nextrt);
}


/*
 * withdraw the selected route, send a status message and do _no_ train
 * number transfer
 */
void Router::withdrawRoute(Route* rt)
{
    if (rt == NULL || visualmode == kvmEditLayout)
        return;

    rt->stopRouting();
    emit statusMessage(tr("Route '%1' withdrawn").arg(rt->getSectionName()));
    // only for debugging purposes:
    // transferTrainNumber(rt);
}


/*
   old error messages
   case kNormal:
   (tr("No normal route found for entry signal '%1'!").arg(s.data()));
   case kDetour:
   (tr("No detour route found for entry signal '%1'!").arg(s.data()));
   case kHelp:
   (tr("No help route found for entry signal '%1'!").arg(s.data()));
   case kShunting:
   (tr("No shunting route found for entry signal '%1'!").arg(s.data()));
   case kShuntingD:
   (tr("No detour shunting route found for entry signal '%1'!").arg(s.data()));
*/
void Router::setRoute(element* el, GbsButtonState cb, GbsButtonState sb)
{
    if (el == NULL)
        return;

    /*check if entry signal is already choosen*/
    if (selectedStartSig == NULL) {
        Route* sr = getUnlockedRouteWithEntrySignal(el, cb, sb);
        if (sr != NULL) {
            selectedStartSig = el;
            lastcb = cb;
            // send signal to gbs to change mouse cursor
            emit startRouteTimer((Route::RouteType)sr->getType());
        }
        else {
            /*TODO: more detailed error message*/
            QApplication::beep();
            /*send cursor time out to gbs*/
            emit routeFunctionFinished();
            emit statusMessage(tr("No matching route found for entry "
                        "signal '%1'").arg(el->getLabelText()));
        }
    }
    /*exit signal button is pressed*/
    else {
        if (lastcb != cb && !((lastcb == kZhsClicked && cb == kZfsClicked)
                    || (cb == kZhsClicked && lastcb == kZfsClicked))) {
            QApplication::beep();
            emit statusMessage(tr("Mixing signal buttons of different"
                        " type is not allowed."));
        }
        else {
            Route* sr = getUnlockedRouteWithExitSignal(el, cb, sb);
            if (sr != NULL)
                activateRoute(sr);
            else {
                QApplication::beep();
                emit statusMessage(tr("No matching route found from '%1'"
                            " to '%2'").arg(selectedStartSig->getLabelText())
                                .arg(el->getLabelText()));
            }
        }
        selectedStartSig = NULL;
        /*send cursor time out to gbs*/
        emit routeFunctionFinished();
    }
}


void Router::resetRoute(element* el, GbsButtonState cb)
{
    if (el == NULL)
        return;

    /*check if route to reset is already choosen*/
    if (resetRt == NULL) {
        resetRt = getLockedRouteWithEntrySignal(el);
        if (resetRt != NULL) {
            lastcb = cb;
            selectedStartSig = el;
            // send signal to gbs to change mouse cursor
            emit startRouteTimer((Route::RouteType)resetRt->getType());
        }
        else {
            QApplication::beep();
            emit statusMessage(tr("No active route found for entry "
                        "signal '%1'").arg(el->getLabelText()));
            emit routeFunctionFinished();
        }
    }
    else {
        if (lastcb != cb) {
            QApplication::beep();
            emit statusMessage(tr("Mixing signal buttons of different"
                        " type is not allowed."));
        }
        else {
            // Check if selected route has same exit signal.
            // If true, do a a FHT withdrawal (no release).
            if (resetRt->hasThisExitSignal(el))
                withdrawRoute(resetRt);
            else {
                QApplication::beep();
                if (selectedStartSig != NULL)
                    emit statusMessage(tr("No active route found from "
                                "entry signal '%1' to exit signal '%2'")
                            .arg(selectedStartSig->getLabelText())
                            .arg(el->getLabelText()));
            }
        }
        /*send cursor time out to gbs*/
        emit routeFunctionFinished();
        resetRt = NULL;
        selectedStartSig = NULL;
    }
}


Route* Router::getReservedRouteWithExitSignal(element* el)
{
    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        if ((*it)->hasThisExitSignal(el) &&
                Route::rsWfLock == (*it)->getState())
            return (*it);
    }
    return NULL;
}


Route* Router::getLockedRouteWithEntrySignal(element* el)
{
    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        if ((*it)->isLockedWithEntrySignal(el))
            return (*it);
    }
    return NULL;
}


Route* Router::getUnlockedRouteWithEntrySignal(element* el, GbsButtonState cb,
        GbsButtonState sb)
{
    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        if ((*it)->isUnlockedWithEntrySignalType(el, cb, sb))
            return (*it);
    }
    return NULL;
}


Route* Router::getUnlockedRouteWithExitSignal(element* el, GbsButtonState cb,
        GbsButtonState sb)
{
    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        if ((*it)->isUnlockedType(selectedStartSig, el, cb, sb))
            return (*it);
    }
    return NULL;
}

/*
 * find route currently used by specified train
 */
Route* Router::getRouteUsedByTrain(unsigned int trainid)
{
    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        if ((*it)->getTrain() == trainid)
            return (*it);
    }
    return NULL;
}

/*
 * find adjacent route with train number
 */
Route* Router::getUtilizedRouteToSignal(element* el)
{
    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        if ((*it)->hasThisExitSignal(el) && (*it)->hasTrain())
            return (*it);
    }
    return NULL;
}

/*
 * find route to same track with train number
 */
Route* Router::getUtilizedRouteToTrack(unsigned int track)
{
    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        if (((*it)->getExitTrack() == track) && (*it)->hasTrain())
            return (*it);
    }
    return NULL;
}

/*
 * find maximum track number for next new route
 * we start with 100 because track numbers have allways three digits
 */
unsigned int Router::getMaximumRouteIdNumber()
{
    unsigned int max = 0;
    unsigned int id = 0;

    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        id = (*it)->getId();
        if (id > max)
            max = id;
    }
    return max;
}


void Router::resetSelectedSignal()
{
    selectedStartSig = NULL;
    resetRt = NULL;
}


void Router::unlockAllLockedRoutes()
{
    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
    int index = 0;
        if ((*it)->getState() != Route::rsUnlocked) {

            // this is a withdrawal, no release
            withdrawRoute(*it);
        }
        ++index;
    }
    emit statusMessage(tr("All active routes withdrawn"));
}


void Router::feedbackPortChanged(unsigned int bus, unsigned int port,
    bool ison)
{
    QList<Route*>::const_iterator it;

    /*first release locked routes*/
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        if ((*it)->canReleaseByFeedbackPort(bus, port, ison))
            releaseRoute(*it);
    }

    /*second activate unlocked routes*/
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        if ((*it)->canActivateByFeedbackPort(bus, port, ison))
            activateRoute(*it);
    }
}


bool Router::editRouteAt(QWidget* owner, int index)
{
    return editRoute(owner, routeList.at(index));
}


bool Router::editRoute(QWidget* owner, Route* er)
{
    bool isEdited = false;

    if (er != NULL) {
        connect(er, SIGNAL(getElementByAddress(const int, const int,
                        element**)), this,
                SIGNAL(getElementByAddress(const int, const int,
                        element**)));
        isEdited = er->runEditDialog(owner);

        disconnect(er, SIGNAL(getElementByAddress(const int,
                        const int, element**)), this,
                SIGNAL(getElementByAddress(const int, const int,
                        element**)));
        if (isEdited)
            modified = true;
    }
    return isEdited;
}

/*
 * set server Generic Messages capabilities
 */
void Router::setServerHasGm(bool gm)
{
    serverhasgm = gm;
}

/*
 * return value for stored server Generic Messages capabilities
 */
bool Router::serverHasGm()
{
    return serverhasgm;
}

/*
 * search for route with specified id
 */
Route* Router::getRouteWithId(unsigned int id)
{
    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        if ((*it)->getId() == id)
            return (*it);
    }
    return NULL;
}

/*
 * set train number of specified route
 */
void Router::changeTrainNumber(unsigned int rn, unsigned int tn)
{
    Route* rt = getRouteWithId(rn);

    if (rt != NULL) {
        rt->setTrain(tn);
        emit routeDataChanged(rt);
        modified = true;
        if (pref.gmbroadcasttrainnumber) {
            QString cms = rt->getCrcfInfoMessage(CrcfMessage::atTrain);
            sendGmCrcfMessage(0, infosessionid, cms);
        }
    }
    else
        emit statusMessage(tr("Error setting train number: "
                    "Route id '%1' not found.").arg(tn));
}

/* Respond to request which route is next for specified train*/
// TRAIN <trainid> GET NEXTROUTE
void Router::handleCrcfTrainMessage(unsigned int replyto, const CrcfMessage* cm)
{
    if (CrcfMessage::meGet != cm->getMethod())
        return;

    if (CrcfMessage::atNextRoute != cm->getAttribute())
        return;

    Route* rt = getRouteUsedByTrain(cm->getActorId());
    if (NULL == rt)
        return;

    Route* nextrt = getLockedRouteWithEntrySignal(
            rt->getExitSignalElementPtr());
    if (NULL == nextrt)
        return;

    QString cms = CrcfMessage::message(CrcfMessage::acTrain,
            cm->getActorId(), CrcfMessage::meInfo,
            CrcfMessage::atNextRoute, nextrt->getId());
    sendGmCrcfMessage(replyto, infosessionid, cms);
}


/*
 * respond to incomming Generic Messages
 */
void Router::processGenericMessage(unsigned int sendto,
        unsigned int replyto, const CrcfMessage* cm)
{
    QString cms = "";

    if (NULL == cm)
        return;

    Route* rt = getRouteWithId(cm->getActorId());
    if (NULL == rt)
        return;

    switch (cm->getMethod()) {
        case CrcfMessage::meGet:
            /* ROUTE <routeid> GET <attribute> */
            cms = rt->getCrcfInfoMessage(cm->getAttribute());
            if (cms.isEmpty())
                emit statusMessage(tr("Unsupported CRCF attribute '%1' "
                            "detected.").arg(cm->getAttributeStr()));
            else
                sendGmCrcfMessage(sendto, infosessionid, cms);
            break;

        case CrcfMessage::meSet:
            /* ROUTE <routeid> SET <attribute> <att_value> */
            switch (cm->getAttribute()) {

                /* ROUTE <routeid> SET STATE <att_value> */
                case CrcfMessage::atState:
                    if (0 == cm->getAttValue() &&
                            rt->getState() != Route::rsUnlocked) {
                        rt->setCrcfMessageReplyId(sendto);
                        withdrawRoute(rt);
                    }
                    else if (1 == cm->getAttValue() &&
                            rt->getState() == Route::rsUnlocked) {
                        rt->setCrcfMessageReplyId(sendto);
                        activateRoute(rt);
                    }
                    else
                        emit statusMessage(tr("Unvalid STATE "
                                    "value '%1' detected.")
                                .arg(cm->getAttValue()));
                    break;

                /* ROUTE <routeid> SET TRAIN <att_value> */
                case CrcfMessage::atTrain:
                    rt->setTrain(cm->getAttValue());
                    emit routeDataChanged(rt);
                    modified = true;
                    cms = rt->getCrcfInfoMessage(CrcfMessage::atTrain);
                    if (!cms.isEmpty()) {
                        if (pref.gmbroadcasttrainnumber)
                            sendGmCrcfMessage(0, replyto, cms);
                        else
                            sendGmCrcfMessage(sendto, replyto, cms);
                    }
                    else
                        emit statusMessage(tr("Error assembling "
                                    "CRCF message for route TRAIN "
                                    "'%1'.").arg(cm->getAttValue()));
                    break;

                /* route type is a static CRCF value, editing is
                 * only allowed in route edit mode */
                /* ROUTE <routeid> SET TYPE <att_value> */
                case CrcfMessage::atType:
                    if (visualmode == kvmEditRoute) {

                        /*check for valid range 0..5*/
                        if (cm->getAttValue() < 6) {
                            rt->setType((Route::RouteType)cm->getAttValue());
                            emit routeDataChanged(rt);
                            modified = true;
                            cms = rt->getCrcfInfoMessage(CrcfMessage::atType);
                            if (!cms.isEmpty())
                                sendGmCrcfMessage(sendto, replyto, cms);
                            else
                                emit statusMessage(tr("Error assembling "
                                            "CRCF message for route TYPE "
                                            "'%1'.").arg(cm->getAttValue()));
                        }
                        else
                            emit statusMessage(tr("Unvalid TYPE "
                                        "value '%1' detected.")
                                    .arg(cm->getAttValue()));
                    }
                    else
                        emit statusMessage(tr("Route type editing via "
                                    "CRCF messages is only allowed in "
                                    "route edit mode."));
                    break;

                /* route id is a static CRCF value, editing is
                 * only allowed in route edit mode */
                /* ROUTE <routeid> SET ID <att_value> */
                case CrcfMessage::atId:
                    if (visualmode == kvmEditRoute) {
                        unsigned int mid = cm->getAttValue();
                        if (mid != rt->getId()) {
                            Route* ir = getRouteWithId(mid);
                            if (NULL == ir) {
                                rt->setId(mid);
                                emit routeDataChanged(rt);
                                modified = true;
                                cms = rt->getCrcfInfoMessage(CrcfMessage::atId);
                                if (!cms.isEmpty())
                                    sendGmCrcfMessage(sendto, replyto, cms);
                                else
                                    emit statusMessage(tr("Error assembling "
                                                "CRCF message for route ID "
                                                "'%1'.").arg(mid));
                            }
                            else
                                emit statusMessage(tr("Route ID '%1' already "
                                            "in use.").arg(mid));
                        }
                    }
                    else
                        emit statusMessage(tr("Route id editing via "
                                    "CRCF messages is only allowed in "
                                    "route edit mode."));
                    break;

                /* route name is a static CRCF value, editing is
                 * only allowed in route edit mode */
                /* ROUTE <routeid> SET NAME <att_value> */
                case CrcfMessage::atName:
                    if (visualmode == kvmEditRoute) {
                        rt->setSectionName(cm->getAttValueStr());
                        emit routeDataChanged(rt);
                        modified = true;
                        cms = rt->getCrcfInfoMessage(CrcfMessage::atName);
                        if (!cms.isEmpty())
                            sendGmCrcfMessage(sendto, replyto, cms);
                        else
                            emit statusMessage(tr("Error assembling "
                                        "CRCF message for route NAME "
                                        "'%1'.").arg(cm->getAttValueStr()));
                    }
                    else
                        emit statusMessage(tr("Route name editing via "
                                    "CRCF messages is only allowed in "
                                    "route edit mode."));
                    break;

                default:
                    break;
            }
            break;

        case CrcfMessage::meInfo:
            /* ROUTE <routeid> INFO <attribute> <att_value> */
            switch (cm->getAttribute()) {

                /* ROUTE <routeid> INFO STATE <att_value> */
                case CrcfMessage::atState:
                    rt->setState((Route::RouteState)cm->getAttValue());
                    break;

                /* ROUTE <routeid> INFO TRAIN <att_value> */
                case CrcfMessage::atTrain:
                    rt->setTrain((Route::RouteState)cm->getAttValue());
                    emit routeDataChanged(rt);
                    modified = true;
                    break;

                default:
                    emit statusMessage(tr("Unsupported CRCF attribute '%1' "
                                "detected.").arg(cm->getAttributeStr()));
                    break;
            }
            break;

        default:
            emit statusMessage(tr("Unsupported CRCF method '%1' "
                        "detected.").arg(cm->getMethodStr()));
            break;
    }
}


/* send a SRCP GM CRCF message to server */
void Router::sendGmCrcfMessage(unsigned int sendto,
        unsigned int replyto, const QString& cms)
{
    GmMessage sm = GmMessage(SrcpMessage::mtGmSet, sendto, replyto,
            "CRCF", cms);
    emit sendSrcpMessage(&sm);
}

void Router::setInfoSessionId(unsigned int id)
{
    infosessionid = id;
}

/* send all available train numbers via CRCF message to server */
void Router::sendTrainNumbers()
{
    QList<Route*>::const_iterator it;
    for (it = routeList.constBegin(); it != routeList.constEnd(); ++it) {
        if ((*it)->hasTrain()) {
            QString cms = (*it)->getCrcfInfoMessage(CrcfMessage::atTrain);
            sendGmCrcfMessage(0, infosessionid, cms);
        }
    }
}
